/*--------------------------------------------------------------
    trackpeerd - peer tracking agent    v0.01(beta)
      Written by  H.Goto, Jan 2005
      Modified by H.Goto, Mar 2005
      http://www.sc.isc.tohoku.ac.jp/~hgot/
--------------------------------------------------------------*/

/*----------------------------------------------------------------------
  Copyright (C) 2005  Hideaki Goto

  All Rights Reserved.

  Permission is hereby granted, free of charge, to any person obtaining
  a copy of this software and associated documentation files (the
  "Software"), to deal in the Software without restriction, including
  without limitation the rights to use, copy, modify, merge, publish,
  distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to
  the following conditions:

  The above copyright notice and this permission notice shall be
  included in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
  IN NO EVENT SHALL THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR
  ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Except as contained in this notice, the name(s) of the copyright
  holder(s) shall not be used in advertising or otherwise to promote
  the sale, use or other dealings in this Software without prior
  written authorization of the copyright holder(s).
----------------------------------------------------------------------*/

#include	"trackpeerd.h"

#include	<signal.h>
#include	<net/ethernet.h>
#include	<netinet/ip.h>
#define	__FAVOR_BSD
#include	<netinet/tcp.h>
#include	<pcap.h>


#define	TDIFF_e	5
#define	TDIFF_l	5
#define	CacheTimeSpan	600	/* 10min */
#define	EventCacheSize	10240
#define	FWtableSize	8
#define	RDtableSize	8

/*
  ALLOWWILDCARDs is defined in the default configuration, since
  information about a source is not always available on servers' sides.
*/
#define	ALLOWWILDCARDs

/*
  Using the following wildcard options is not recommended for privacy
  and/or security reasons.
*/
#undef	ALLOWWILDCARDd
#define	ALLOWWILDCARDt


#define	PCAPTIMEOUT	5
#define	IFNAMELEN	10

static char	capt_deffilter[] = "tcp[13] & 2 != 0";
static char	cmdverstr[] = CMDVERSTRING;
static uint16_t	agentport = TMAC_PORT;


typedef struct {
	time_t		evtime;		/* time (sec) of event occurence in UTC */
	in_addr_t	ip_src;
	uint16_t	port_src;
	struct ether_addr	mac_src;
	in_addr_t	ip_dst;
	uint16_t	port_dst;
	struct ether_addr	mac_dst;
	uint16_t	port_src_xlat;
	int		natmode;	/* 0: no NAT, 1: NAT+table, 2: NAT w/o table */
} EventCache_item;


typedef struct {
	int		ttl;
	time_t		evtime;		/* time (sec) of event occurence in UTC */
	in_addr_t	ip_src;
	uint16_t	port_src;
	in_addr_t	ip_dst;
	uint16_t	port_dst;
	in_addr_t	ip_client;
	long long	serial;
} QueryInfo;


typedef struct {
	in_addr_t	ip_src;
	in_addr_t	mask_src;
	in_addr_t	ip_next;
	uint16_t	port_next;
} FWTBL;


/* flags and options */
static int	f_debug = 0;
static int	f_nat = 0;
static int	f_promisc = 0;
static char	*capt_filter = capt_deffilter;
static char	capt_dev[IFNAMELEN+1] = "";
//static char	capt_dev[IFNAMELEN+1] = "eth0";
static char	*conffile = "/etc/trackpeerd.conf";

/* flags for internal use */
static int	f_terminate = 0;
static int	f_pcap = 0;
static int	f_pcap_status = 0;	/* 0:not ready, 1:ready, 2:error */

static pthread_mutex_t	mtx;	/* mutex */

		/* Ring buffer for event caching */
static EventCache_item	EventCache[ EventCacheSize ];
static int	ec_count = 0;
static int	ec_current = -1;

		/* Event buffer for data transmission */
static EventCache_item	EventCache_buf[ EventCacheSize ];

		/* Table for query forwarding */
static int	fw_count = 0;
static FWTBL	FWtable[ FWtableSize ];

		/* Table for query redirection */
static int	rd_count = 0;
static FWTBL	RDtable[ RDtableSize ];




/*--------------------------------------------------------------
    Signal handler
--------------------------------------------------------------*/

void signal_handler(int sigtype){
	/* ignore SIGPIPE */
//	fprintf(stderr,"Caught SIGPIPE.\n");
}




/*--------------------------------------------------------------
    NAT table look-up
      Try to get port_src after the translation,
      and set natmode flag according to the environment.
--------------------------------------------------------------*/

static void lookup_nattable(EventCache_item *ev){
	ev->natmode = 2;
	ev->port_src_xlat = ev->port_src;
	if ( !f_nat ){	/* no NAT */
		ev->natmode = 0;
		return;
	}

/*
	NAT table look-up code should be here.
	It hasn't been implemented yet.
	The code would be strongly dependent on the operating system used.
*/

}




/*--------------------------------------------------------------
    Ring buffer handling
--------------------------------------------------------------*/

static int ec_putnewitem(EventCache_item *item){
	int	i;
	int	skip = 0;
	int	pos;
	pthread_mutex_lock(&mtx);
	pos = ec_current;
	for ( i=0 ; i<ec_count ; i++ ){
		if ( item->evtime != EventCache[pos].evtime )  break;
		if ( item->ip_src == EventCache[pos].ip_src \
		  && item->port_src == EventCache[pos].port_src \
		  && item->ip_dst == EventCache[pos].ip_dst \
		  && item->port_dst == EventCache[pos].port_dst ){
			skip = 1;  break;
		}
		if ( --pos < 0 )  pos += EventCacheSize;
	}
	if ( skip ){	/* avoid creating duplicate entries */
		pthread_mutex_unlock(&mtx);
		return(0);
	}

	if ( ++ec_current >= EventCacheSize )  ec_current = 0;

	lookup_nattable(item);

	EventCache[ ec_current ] = *item;
	if ( ec_count < EventCacheSize )  ec_count++;
	pthread_mutex_unlock(&mtx);
	return(1);
}




/*--------------------------------------------------------------
    Packet capturing thread
--------------------------------------------------------------*/

void capt_callback(u_char *usrdata, const struct pcap_pkthdr *h, const u_char *p){
	struct ether_header	*p_ether;
	struct ip	*p_ip;
	struct tcphdr	*p_tcp;
	EventCache_item	ev;
	char	eaddrstr_s[20];
	char	eaddrstr_d[20];
	char	addrstr_s[20];
	char	addrstr_d[20];

	/* skip packets that are too short */
	if ( h->len < sizeof(struct ether_header) + sizeof(struct ip) \
		+ sizeof(struct tcphdr) )  return;

	p_ether = (struct ether_header *)p;
	p_ip = (struct ip *)(p + sizeof(struct ether_header));
	p_tcp = (struct tcphdr *)(p + sizeof(struct ether_header) + sizeof(struct ip));

	/* skip non-IP packets */
	if ( ntohs(p_ether->ether_type) != ETHERTYPE_IP )  return;

	ev.evtime = h->ts.tv_sec;
	memcpy((void *)&ev.mac_src, \
		(void *)p_ether->ether_shost,sizeof(struct ether_addr));
	memcpy((void *)&ev.mac_dst, \
		(void *)p_ether->ether_dhost,sizeof(struct ether_addr));
	ev.ip_src = p_ip->ip_src.s_addr;
	ev.ip_dst = p_ip->ip_dst.s_addr;
	ev.port_src = ntohs(p_tcp->th_sport);
	ev.port_dst = ntohs(p_tcp->th_dport);

	ec_putnewitem(&ev);

	if ( f_debug > 1 ){
		strcpy(eaddrstr_s,ether_ntoa(&ev.mac_src));
		strcpy(eaddrstr_d,ether_ntoa(&ev.mac_dst));
		strcpy(addrstr_s,inet_ntoa(p_ip->ip_src));
		strcpy(addrstr_d,inet_ntoa(p_ip->ip_dst));
		printf("%lld  %s %s:%hd > %s %s:%hd  (%d)\n", \
			(long long)ev.evtime, \
			eaddrstr_s, addrstr_s, ev.port_src, \
			eaddrstr_d, addrstr_d, ev.port_dst, ec_count);
		fflush(stdout);
	}
}


void *pcap_thread(void *pargs){
	pcap_t	*pd;
	int	snaplen = 64;
	struct bpf_program	fcode;
	bpf_u_int32	localnet, netmask;
	char	ebuf[PCAP_ERRBUF_SIZE];

	printf("Starting packet captor: dev=%s filter='%s'\n", \
		capt_dev, capt_filter);

	if ( NULL == (pd = pcap_open_live(capt_dev, snaplen, \
		f_promisc, 1000, ebuf)) ){
		fputs("Packet captor (libpcap) initialization failed.\n",stderr);
		f_pcap_status = 2;
		return NULL;
	}

	if ( 0 > pcap_lookupnet(capt_dev, &localnet, &netmask, ebuf) ){
		fputs("Can't get the interface information.\n",stderr);
		f_pcap_status = 2;
		return NULL;
	}
//printf("localnet=%08X  netmask=%08X\n",(int)ntohl((uint32_t)localnet),(int)ntohl((uint32_t)netmask));

	if ( 0 > pcap_compile(pd, &fcode, capt_filter, 1, netmask) ){
		fputs("Can't compile the filter.\n",stderr);
		f_pcap_status = 2;
		return NULL;
	}
	if ( 0 > pcap_setfilter(pd, &fcode) ){
		fputs("Can't set the filter.\n",stderr);
		f_pcap_status = 2;
		return NULL;
	}

	f_pcap_status = 1;
	if ( 0 > pcap_loop(pd, -1, capt_callback, NULL) ){
		fputs("Failed on packet capturing.\n",stderr);
		f_pcap_status = 2;
		return NULL;
	}

	pcap_close(pd);
	fputs("Packet captor (libpcap) has died.\n",stderr);
	f_pcap_status = 2;
	f_terminate = 1;
	return NULL;
}




/*--------------------------------------------------------------
    Forward query
--------------------------------------------------------------*/

static int fwdquery(int clSock, \
	in_addr_t ip_next, uint16_t port_next, QueryInfo *qi){
	int	svSock;
	struct sockaddr_in	svAddr;
	char	msgbuf[MAXMSGLEN];
	char	querystr[MAXMSGLEN];
	char	rescode[MAXMSGLEN];
	char	addrstr[30];
	int	ttl;

/* connect to the next agent */
	memset(&svAddr, 0, sizeof(svAddr));
	svAddr.sin_family = AF_INET;
	svAddr.sin_addr.s_addr = ip_next;
	svAddr.sin_port = htons(port_next);

	if ( 0 > (svSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) ){
		return(1);	/* connection failed */
	}
	if ( 0 > connect(svSock, (struct sockaddr *)&svAddr, sizeof(svAddr)) ){
		close(svSock);
		return(1);	/* connection failed */
	}

	/* greeting */
	if ( 0 > SendLineLF(svSock, GREETINGSTR, SENDTIMEOUT) ){
		close(svSock);
		return(2);	/* timeout (next agent) */
	}
	if ( 0 > ReceiveLine(svSock, msgbuf, MAXMSGLEN, REPLYTIMEOUT) ){
		close(svSock);
		return(2);	/* timeout (next agent) */
	}
	if ( strcmp(msgbuf, CMDVERSTRING"\n") ){
		close(svSock);
		return(4);	/* command version mismatch */
	}


/* format the query command */
	sprintf(querystr,"FIND %d  %lld ", \
		qi->ttl, (long long)qi->evtime);
	inet_ntop(AF_INET,&qi->ip_src,addrstr,INET_ADDRSTRLEN);
	strcat(querystr,addrstr);
	sprintf(addrstr," %d ",(int)((uint)qi->port_src));
	strcat(querystr,addrstr);
	inet_ntop(AF_INET,&qi->ip_dst,addrstr,INET_ADDRSTRLEN);
	strcat(querystr,addrstr);
	sprintf(addrstr," %d  ",(int)((uint)qi->port_dst));
	strcat(querystr,addrstr);
	inet_ntop(AF_INET,&qi->ip_client,addrstr,INET_ADDRSTRLEN);
	strcat(querystr,addrstr);
	sprintf(addrstr," %lld",qi->serial);
	strcat(querystr,addrstr);

/* issue the command */
	if ( 0 > SendLineLF(svSock, querystr, SENDTIMEOUT) ){
		close(svSock);
		return(2);	/* timeout (next agent) */
	}

/* receive result lines */
	for (;;){
		if ( 0 > ReceiveLine(svSock, msgbuf, MAXMSGLEN, REPLYTIMEOUT) ){
			close(svSock);
			return(2);	/* timeout (next agent) */
		}
		if ( 0 > SendLine(clSock, "  ", SENDTIMEOUT) ){
			close(svSock);
			return(3);	/* timeout (client) */
		}
		if ( 0 > SendLine(clSock, msgbuf, SENDTIMEOUT) ){
			close(svSock);
			return(3);	/* timeout (client) */
		}
		ttl = qi->ttl;
		sscanf(msgbuf,"%s %d",rescode,&ttl);

		/* exit upon fatal error */
		if ( 0 == strcmp(rescode,"CMDERR") )  break;
		if ( 0 == strcmp(rescode,"END") && ttl == qi->ttl )  break;
	}


/* disconnect from the next agent */
	if ( 0 > SendLineLF(svSock, "BYE", SENDTIMEOUT) ){
		close(svSock);
		return(2);	/* timeout (next agent) */
	}

	close(svSock);
	return(0);
}




/*--------------------------------------------------------------
    Command handler
--------------------------------------------------------------*/

static int SendStatus(int sock, char *rescode, int ttl){
	char	msgbuf[MAXMSGLEN];
	sprintf(msgbuf, "%s %d", rescode, ttl);
	return( SendLineLF(sock, msgbuf, SENDTIMEOUT) );
}


static int pick_matched(EventCache_item *matched_event, \
	time_t evtime, struct ether_addr *mac_src, \
	in_addr_t ip_src, uint16_t port_src, \
	in_addr_t ip_dst, uint16_t port_dst){
	int	i,pos;
	int	n = 0;
	long long	tdiff;
	time_t	currtime;
	EventCache_item	*ev;
	time(&currtime);
	if ( evtime == -1 ){	/* -1: use current time */
		evtime = currtime;
	}
	pthread_mutex_lock(&mtx);
	pos = ec_current;
	for ( i=0 ; i<ec_count ; i++ ){
		ev = &EventCache[pos];

		/* forget old events */
		if ( CacheTimeSpan < (long long)currtime - (long long)ev->evtime ){
			ec_count = i;
			break;
		}

		if ( --pos < 0 )  pos += EventCacheSize;
		if ( evtime != 0 ){	/* 0: pick all */
			tdiff = (long long)evtime - (long long)ev->evtime;
			if ( tdiff >  TDIFF_e )  continue;
			if ( tdiff < -TDIFF_l )  continue;
		}
		if ( ip_dst != ip_all ){
			if ( ip_dst != ev->ip_dst )  continue;
		}
		if ( port_dst != port_all ){
			if ( port_dst != ev->port_dst )  continue;
		}

		if ( ev->natmode == 0 ){	/* no NAT */
			if ( ip_src != ip_all ){
				if ( ip_src != ev->ip_src )  continue;
			}
			if ( port_src != port_all ){
				if ( port_src != ev->port_src )  continue;
			}
		}
		if ( ev->natmode == 1 ){	/* NAT + mapping table */
			if ( port_src != port_all ){
				if ( port_src != ev->port_src_xlat )  continue;
			}
		}

		matched_event[n++] = *ev;
	}
	pthread_mutex_unlock(&mtx);
	return(n);
}


static int parse_querystr(QueryInfo *qi, char *cmd){
	int	retcode = 0;
	char	prm[9][MAXMSGLEN];

	prm[8][0] = '\0';
	sscanf(cmd,"%s %s %s %s %s %s %s %s %s", \
		prm[0], prm[1], prm[2], prm[3], \
		prm[4], prm[5], prm[6], prm[7], \
		prm[8]);

/* FIND ttl time ip_src port_src ip_dst port_dst ip_client serial */
/*  0    1   2    3      4        5      6        7         8     */

	/* check parameters */
	if ( 0 == strlen(prm[8]) ){	/* correct parameter length? */
		return(1);	/* command error */
	}
	qi->ttl = atoi(prm[1]);
	if ( qi->ttl > MAXTTL )  qi->ttl = MAXTTL;
	if ( -1 > atoll(prm[1]) )  retcode = 2;
	qi->evtime = (time_t)atoll(prm[2]);
	qi->ip_src = qi->ip_dst = ip_all;
	if ( 0 >= inet_pton(AF_INET,prm[3],&qi->ip_src) )  retcode = 2;
	qi->port_src = (uint16_t)((unsigned int)atoi(prm[4]));
	if ( 0 >= inet_pton(AF_INET,prm[5],&qi->ip_dst) )  retcode = 2;
	qi->port_dst = (uint16_t)((unsigned int)atoi(prm[6]));
	qi->ip_client = ip_null;
	if ( 0 >= inet_pton(AF_INET,prm[7],&qi->ip_client) )  retcode = 2;
	qi->serial = atoll(prm[8]);
	return(retcode);
}


static int cmd_FIND(int clSock, struct sockaddr_in *clAddr, char *cmd){
	int	i,n_matched;
	int	paramerr = 0;
	char	msgbuf[MAXMSGLEN];
	char	addrstr[30];
	char	mflag;
	QueryInfo	qi;
	EventCache_item	*ev;
	in_addr_t	addr;
	struct ether_addr	mac_src;
	int		cttl;
	long long	evtime;
	int	retcode;

	paramerr = parse_querystr(&qi, cmd);
	if ( paramerr ){
		if ( 0 > SendLineLF(clSock, "CMDERR", SENDTIMEOUT) )  return(-1);
		return(0);
	}

	if ( qi.ip_client == ip_null ){	/* insert client address */
		qi.ip_client = clAddr->sin_addr.s_addr;
	}

	cttl = qi.ttl;
	if ( cttl <= 0 ){
		if ( 0 > SendStatus(clSock, "EXPIRED", cttl) )  return(-1);
		return(0);
	}

#ifndef	ALLOWWILDCARDs
//	if ( !f_nat && (qi.ip_src == ip_all || qi.port_src == port_all) ){
	if ( !f_nat && qi.ip_src == ip_all ){
		if ( 0 > SendStatus(clSock, "FORBIDDEN", cttl) )  return(-1);
		goto skiplocal;
	}
#endif	/* ALLOWWILDCARDs */
#ifndef	ALLOWWILDCARDd
	if ( qi.ip_dst == ip_all || qi.port_dst == port_all ){
		if ( 0 > SendStatus(clSock, "FORBIDDEN", cttl) )  return(-1);
		goto skiplocal;
	}
#endif	/* ALLOWWILDCARDd */
#ifndef	ALLOWWILDCARDt
	if ( 0 == qi.evtime ){
		if ( 0 > SendStatus(clSock, "FORBIDDEN", cttl) )  return(-1);
		goto skiplocal;
	}
#endif	/* ALLOWWILDCARDt */


	n_matched = pick_matched(EventCache_buf, qi.evtime, &mac_src, \
		qi.ip_src, qi.port_src, qi.ip_dst, qi.port_dst);


/* for local event handling */

	if ( n_matched == 0 ){
		if ( 0 > SendStatus(clSock, "NOTFOUND", cttl) )  return(-1);
	}

	for ( i=0 ; i<n_matched ; i++ ){
		ev = &EventCache_buf[i];
		evtime = (long long)ev->evtime;
		ether_ntoa_r(&ev->mac_src,addrstr);
		mflag = 'E';
		if ( ev->natmode > 1 )  mflag = 'L';
		sprintf(msgbuf,"FOUND %d %lld %c %s ", \
			cttl, (long long)evtime, mflag, addrstr);
		inet_ntop(AF_INET,&ev->ip_src,addrstr,INET_ADDRSTRLEN);
		strcat(msgbuf,addrstr);
		sprintf(addrstr," %d ",(int)((uint)ev->port_src));
		strcat(msgbuf,addrstr);
		inet_ntop(AF_INET,&ev->ip_dst,addrstr,INET_ADDRSTRLEN);
		strcat(msgbuf,addrstr);
		sprintf(addrstr," %d ",(int)((uint)ev->port_dst));
		strcat(msgbuf,addrstr);
		inet_ntop(AF_INET,&qi.ip_client,addrstr,INET_ADDRSTRLEN);
		strcat(msgbuf,addrstr);
		sprintf(addrstr," %lld", qi.serial);
		strcat(msgbuf,addrstr);
		if ( 0 > SendLineLF(clSock, msgbuf, SENDTIMEOUT) )  return(-1);
	}


skiplocal:

/* for query forwarding/redirection */

	qi.ttl--;
	for ( i=0 ; i < rd_count ; i++ ){
		if ( qi.ip_src != ip_all ){
			addr = qi.ip_src & RDtable[i].mask_src;
			if ( addr != (RDtable[i].ip_src & RDtable[i].mask_src) )  continue;
		}
		sprintf(msgbuf,"REDIRECTED %d ",cttl);
		inet_ntop(AF_INET,&RDtable[i].ip_next,addrstr,INET_ADDRSTRLEN);
		strcat(msgbuf,addrstr);
		sprintf(addrstr," %d",(int)((uint)RDtable[i].port_next));
		strcat(msgbuf,addrstr);
		if ( 0 > SendLineLF(clSock, msgbuf, SENDTIMEOUT) )  return(-1);
		if ( qi.ttl < 1 ){
			/* The expiration message is generated by the */
			/* current agent instead of the next agent to */
			/* reduce useless communications.             */
			sprintf(msgbuf,"EXPIRED %d",qi.ttl);
			if ( 0 > SendLineLF(clSock, msgbuf, SENDTIMEOUT) )  return(-1);
			break;
		}

		retcode = fwdquery(clSock, RDtable[i].ip_next, RDtable[i].port_next, &qi);
		switch(retcode){
 		  case 1:	/* connection failed */
			sprintf(msgbuf,"NOROUTE %d ",cttl);
			if ( 0 > SendLineLF(clSock, msgbuf, SENDTIMEOUT) )  return(-1);
			break;
		  case 2:	/* timeout (next agent) */
			break;
		  case 3:	/* timeout (client) */
			return(-1);
		  case 4:	/* command mismatch */
			sprintf(msgbuf,"PROTOERR %d ",cttl);
			if ( 0 > SendLineLF(clSock, msgbuf, SENDTIMEOUT) )  return(-1);
			break;
		}
	}

/*
	A query forwarding code may be added here.
*/


	/* end of the session */
	if ( 0 > SendStatus(clSock, "END", cttl) )  return(-1);

	return(0);
}


static void cmdhandler(int clSock, struct sockaddr_in *clAddr){
	char	buf[MAXMSGLEN];
	char	cmd[MAXMSGLEN];
	int	rv;
	int	f_engaged = 0;
	for (;;){
		rv = ReceiveLine(clSock, buf, MAXMSGLEN, CMDTIMEOUT);
		if ( 0 > rv ){
			fputs("Warning: Command transmission error.\n",stderr);
			return;
		}
		if ( 0 == rv ){
			fputs("Warning: Timeout occurred (recv).\n",stderr);
			return;
		}

		cmd[0] = '\0';
		sscanf(buf,"%s",cmd);

		if ( strcmp(cmd,"trackpeer") == 0 ){	/* gives command version */
			if ( 0 > SendLineLF(clSock, cmdverstr, SENDTIMEOUT) )  break;
			f_engaged = 1;	/* accept other commands afterward */
			continue;
		}

		/* Disconnect, if there was no greeting. */
		if ( ! f_engaged ){
			fputs("Warning: No greeting.\n",stderr);
			return;
		}

		if ( strcmp(cmd,"BYE") == 0 )  break;

		if ( strcmp(cmd,"FIND") == 0 ){
			if ( 0 > cmd_FIND(clSock, clAddr, buf) )  return;
			continue;
		}

		if ( 0 > SendLineLF(clSock, "CMDERR", SENDTIMEOUT) )  return;
	}
}




/*--------------------------------------------------------------
    Read configuration file
--------------------------------------------------------------*/

#define	CONFFILE_LINELEN	256

static int readconf(char *fname){
	FILE	*fp;
	char	lbuf[ CONFFILE_LINELEN+1 ];
	char	cmd[ CONFFILE_LINELEN+1 ];
	char	prm1[ CONFFILE_LINELEN+1 ];
	char	prm2[ CONFFILE_LINELEN+1 ];
	char	*p;
	int	n;
	in_addr_t	addr,mask;
	uint16_t	port;
	if ( NULL == (fp = fopen(fname,"r")) )  return(1);
	while ( NULL != (p = fgets(lbuf, CONFFILE_LINELEN, fp)) ){
		p = strchr(lbuf,'#');
		if ( NULL != p )  *p = '\0';
		cmd[0] = '\0';
		prm1[0] = prm2[0] = '\0';
		sscanf(lbuf,"%s %s %s",cmd,prm1,prm2);
		if ( 0 == strlen(cmd) )  continue;
		if ( 0 == strcasecmp(cmd,"agentport") ){
			if ( prm1[0] != '\0' ){
				n = atoi(prm1);
				if ( n > 0 && n < 65535 )
					agentport = (uint16_t)((unsigned int)n);
			}
		}
		else if ( 0 == strcasecmp(cmd,"listen") ){
			if ( prm1[0] != '\0' \
			  && IFNAMELEN > strlen(prm1) ){
				strcpy(capt_dev, prm1);
			}
		}
		else if ( 0 == strcasecmp(cmd,"forward") ){
			if ( prm1[0] == '\0' || prm2[0] == '\0' )  continue;
			if ( fw_count >= FWtableSize )  continue;
			if ( get_addr_mask(prm1,&addr,&mask) )  continue;
			FWtable[fw_count].ip_src = addr;
			FWtable[fw_count].mask_src = mask;
			if ( get_addr_port(prm2,&addr,&port) )  continue;
			FWtable[fw_count].ip_next = addr;
			FWtable[fw_count].port_next = port;
			fw_count++;
		}
		else if ( 0 == strcasecmp(cmd,"redirect") ){
			if ( prm1[0] == '\0' || prm2[0] == '\0' )  continue;
			if ( rd_count >= RDtableSize )  continue;
			if ( get_addr_mask(prm1,&addr,&mask) )  continue;
			RDtable[rd_count].ip_src = addr;
			RDtable[rd_count].mask_src = mask;
			if ( get_addr_port(prm2,&addr,&port) )  continue;
			RDtable[rd_count].ip_next = addr;
			RDtable[rd_count].port_next = port;
			rd_count++;
		}
	}
	fclose(fp);
	return(0);
}




/*--------------------------------------------------------------
    Help message
--------------------------------------------------------------*/

static void usage(){
	fputs("trackpeerd - peer tracking agent  v0.01(beta)\n",stdout);
	fputs("Copyright (C) 2005 Hideaki Goto\n",stdout);
	fputs("usage: trackpeerd [-options]\n",stdout);
	fputs("    -h            : show help message\n",stdout);
	fputs("    -p            : put the device into promiscuous mode\n",stdout);
	fputs("    -nat          : NAT mode\n",stdout);
	fputs("    -dev DEV      : device to listen\n",stdout);
	fputs("    --pcap FILTER : filter rule for packet captor (libpcap)\n",stdout);
}




/*--------------------------------------------------------------
    Main routine
--------------------------------------------------------------*/

int main(int ac, char *av[]){
	int	n,i,k;
	pthread_t	tid_pcap;
	int		svSock;
	int		clSock;
	unsigned int	clAddrLen;
	struct sockaddr_in	svAddr;
	struct sockaddr_in	clAddr;
	struct sigaction	handler;

	for ( k=1, i=1 ; i<ac && k ; i++ ){
		if ( 0 == strcmp("-h",av[i]) || 0 == strcmp("-help",av[i]) ){
			usage();  return(0);
		}
		if ( 0 == strcmp("-f",av[i]) ){
			if ( i++ >= ac ){  k=0; continue; }
			conffile = av[i];
			continue;
		}
	}
	readconf(conffile);

	for ( k=1, i=1 ; i<ac && k ; i++ ){
		if ( 0 == strcmp("-d",av[i]) ){  f_debug = 1;  continue; }
		if ( 0 == strcmp("-d1",av[i]) ){  f_debug = 1;  continue; }
		if ( 0 == strcmp("-d2",av[i]) ){  f_debug = 2;  continue; }
		if ( 0 == strcmp("-nat",av[i]) ){  f_nat = 1;  continue; }
		if ( 0 == strcmp("-p",av[i]) ){  f_promisc = 1;  continue; }
		if ( 0 == strcmp("-dev",av[i]) ){
			if ( i++ >= ac \
			  || IFNAMELEN < strlen(av[i]) ){  k=0; continue; }
			strcpy(capt_dev, av[i]);
			continue;
		}
		if ( 0 == strcmp("--pcap",av[i]) ){
			if ( i++ >= ac ){  k=0; continue; }
			capt_filter = av[i];
			continue;
		}
		if ( 0 == strcmp("-h",av[i]) || 0 == strcmp("-help",av[i]) ){
			continue;
		}
		if ( 0 == strcmp("-f",av[i]) ){
			if ( i++ >= ac )  k=0;
			continue;
		}

		k = 0;
	}
	if ( k == 0 ){  usage();  return(0); }

	if ( f_debug == 0 ){
		if ( fork() )  exit(0);
	}

	handler.sa_handler = signal_handler;
	sigfillset(&handler.sa_mask);
	handler.sa_flags = 0;
	sigaction(SIGPIPE, &handler, 0);


	/* start packet capturing thread */
	pthread_mutex_init(&mtx, NULL);
	if ( capt_dev[0] != '\0' )  f_pcap = 1;

	if ( f_pcap ){
		pthread_create(&tid_pcap,NULL,pcap_thread,(void *)NULL);

		for ( n=0 ; n<PCAPTIMEOUT ; n++ ){	/* wait for libpcap to start-up */
			sleep(1);
			if ( f_pcap_status )  break;
		}
		if ( 1 != f_pcap_status ){	/* stop the program upon timeout or error */
			return(1);
		}
	}

	/* start command handler */
	if ( 0 > (svSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) ){
		fputs("socket() failed.\n",stderr);
		return(1);
	}

	memset(&svAddr, 0, sizeof(svAddr));
	svAddr.sin_family = AF_INET;
	svAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	svAddr.sin_port = htons(agentport);

	if ( 0 > bind(svSock, (struct sockaddr *)&svAddr, sizeof(svAddr)) ){
		fprintf(stderr,"bind() failed for port %u.\n",(uint)agentport);
		return(1);
	}
	if ( 0 > listen(svSock, MAXPENDING) ){
		fputs("listen() failed.\n",stderr);
		return(1);
	}

	while( ! f_terminate ){
		clAddrLen = sizeof(clAddr);
		if ( 0 > (clSock = accept(svSock, \
			(struct sockaddr *)&clAddr, &clAddrLen)) ){
			fputs("accept() failed.\n",stderr);
			close(clSock);
			continue;
		}

		cmdhandler(clSock,&clAddr);

		close(clSock);
	}

	close(svSock);
	if ( f_pcap )  pthread_mutex_destroy(&mtx);
	return(0);
}

